import canvasapi.file
import canvasapi.submission

import canvaslms.cli.assignments as assignments
import canvaslms.cli.users as users
import canvaslms.hacks.canvasapi

import argparse
import csv
import json
import os
import pathlib
import subprocess
import pypandoc
import re
import rich.console
import rich.markdown
import rich.json
import sys
import tempfile
import urllib.request


def submissions_command(config, canvas, args):
    assignment_list = assignments.process_assignment_option(canvas, args)
    if args.ungraded:
        submissions = list_ungraded_submissions(assignment_list)
    else:
        submissions = list_submissions(assignment_list)

    if args.user or args.category or args.group:
        user_list = users.process_user_or_group_option(canvas, args)
        submissions = filter_submissions(submissions, user_list)

    output = csv.writer(sys.stdout, delimiter=args.delimiter)

    for submission in submissions:
        if args.login_id:
            output.writerow(format_submission_short_unique(submission))
        else:
            output.writerow(format_submission_short(submission))


def speedgrader(submission):
    """Returns the SpeedGrader URL of the submission"""
    speedgrader_url = submission.preview_url

    speedgrader_url = re.sub(
        "assignments/", "gradebook/speed_grader?assignment_id=", speedgrader_url
    )

    speedgrader_url = re.sub("/submissions/", "&student_id=", speedgrader_url)

    speedgrader_url = re.sub(r"\?preview.*$", "", speedgrader_url)

    return speedgrader_url


def submission_command(config, canvas, args):
    submission_list = process_submission_options(canvas, args)
    console = rich.console.Console()
    if args.output_dir:
        tmpdir = pathlib.Path(args.output_dir)
    else:
        tmpdir = pathlib.Path(tempfile.mkdtemp())
    for submission in submission_list:
        if args.sort_order == "student":
            subdir = (
                f"{submission.user.login_id}"
                f"/{submission.assignment.course.course_code}"
                f"/{submission.assignment.name}"
            )
        else:
            subdir = (
                f"{submission.assignment.course.course_code}"
                f"/{submission.assignment.name}"
                f"/{submission.user.login_id}"
            )

        (tmpdir / subdir).mkdir(parents=True, exist_ok=True)
        output = format_submission(
            submission,
            history=args.history,
            tmpdir=tmpdir / subdir,
            json_format=args.json,
        )

        if args.json:
            filename = "metadata.json"
            output = json.dumps(output, indent=2)
        else:
            filename = "metadata.md"
        with open(tmpdir / subdir / filename, "w") as f:
            f.write(output)

        if args.output_dir:
            pass
        elif args.open == "shell":
            print(f"---> Opening a shell ({os.environ['SHELL']}) in {tmpdir/subdir}")
            subprocess.run(
                ["sh", "-c", f"cd '{tmpdir/subdir}' && exec {os.environ['SHELL']}"]
            )
            print(f"<--- canvaslms submission shell terminated.")
        elif sys.stdout.isatty():
            pager = ""
            if "MANPAGER" in os.environ:
                pager = os.environ["MANPAGER"]
            elif "PAGER" in os.environ:
                pager = os.environ["PAGER"]

            styles = False
            if "less" in pager and ("-R" in pager or "-r" in pager):
                styles = True
            if args.open == "open":
                subprocess.run(["open", tmpdir / subdir])
            with console.pager(styles=styles):
                if args.json:
                    console.print(rich.json.JSON(output))
                else:
                    console.print(rich.markdown.Markdown(output, code_theme="manni"))
        else:
            print(output)


def add_submission_options(parser, required=False):
    try:
        assignments.add_assignment_option(parser, required=required)
    except argparse.ArgumentError:
        pass

    try:
        users.add_user_or_group_option(parser, required=required)
    except argparse.ArgumentError:
        pass

    submissions_parser = parser.add_argument_group("filter submissions")
    try:  # to protect from this option already existing in add_assignment_option
        submissions_parser.add_argument(
            "-U", "--ungraded", action="store_true", help="Only ungraded submissions."
        )
    except argparse.ArgumentError:
        pass


def process_submission_options(canvas, args):
    assignment_list = assignments.process_assignment_option(canvas, args)
    user_list = users.process_user_or_group_option(canvas, args)

    if args.ungraded:
        submissions = list_ungraded_submissions(
            assignment_list,
            include=["submission_history", "submission_comments", "rubric_assessment"],
        )
    else:
        submissions = list_submissions(
            assignment_list,
            include=["submission_history", "submission_comments", "rubric_assessment"],
        )

    return list(filter_submissions(submissions, user_list))


def list_submissions(assignments, include=["submission_comments"]):
    for assignment in assignments:
        submissions = assignment.get_submissions(include=include)
        for submission in submissions:
            submission.assignment = assignment
            yield submission


def list_ungraded_submissions(assignments, include=["submisson_comments"]):
    for assignment in assignments:
        submissions = assignment.get_submissions(bucket="ungraded", include=include)
        for submission in submissions:
            if submission.submitted_at and (
                submission.graded_at is None
                or not submission.grade_matches_current_submission
            ):
                submission.assignment = assignment
                yield submission


def filter_submissions(submission_list, user_list):
    user_list = set(user_list)

    for submission in submission_list:
        for user in user_list:
            if submission.user_id == user.id:
                submission.user = user
                yield submission
                break


def format_submission_short(submission):
    return [
        submission.assignment.course.course_code,
        submission.assignment.name,
        submission.user.name,
        submission.grade,
        submission.submitted_at,
        submission.graded_at,
    ]


def format_submission_short_unique(submission):
    uid = users.get_uid(submission.user)

    return [
        submission.assignment.course.course_code,
        submission.assignment.name,
        uid,
        submission.grade,
        submission.submitted_at,
        submission.graded_at,
    ]


def format_submission(
    submission,
    history=False,
    json_format=False,
    md_title_level="#",
    tmpdir=None,
):
    """
    Formats submission for printing to stdout. Returns a string.

    If history is True, include all submission versions from history.

    If json_format is True, return a JSON string, otherwise Markdown.

    `md_title_level` is the level of the title in Markdown, by default `#`. This
    is used to create a hierarchy of sections in the output.

    `tmpdir` is the directory to store all the submission files. Defaults to None,
    which creates a temporary directory.
    """
    student = submission.assignment.course.get_user(submission.user_id)

    if json_format:
        formatted_submission = {}
    else:
        formatted_submission = ""

    metadata = {
        "course": submission.assignment.course.course_code,
        "assignment": submission.assignment.name,
        "student": str(student),
        "submission_id": submission.id,
        "submitted_at": submission.submitted_at,
        "graded_at": submission.graded_at,
        "grade": submission.grade,
        "graded_by": str(resolve_grader(submission)),
        "speedgrader": speedgrader(submission),
    }

    if json_format:
        formatted_submission.update(
            format_section(
                "metadata", metadata, json_format=True, md_title_level=md_title_level
            )
        )
    else:
        formatted_submission += format_section(
            "Metadata", metadata, md_title_level=md_title_level
        )
    try:
        if submission.rubric_assessment:
            if json_format:
                formatted_submission.update(
                    format_section(
                        "rubric_assessment", format_rubric(submission, json_format=True)
                    ),
                    json_format=True,
                )
            else:
                formatted_submission += format_section(
                    "Rubric assessment", format_rubric(submission)
                )
    except AttributeError:
        pass
    try:
        if submission.submission_comments:
            if json_format:
                formatted_submission.update(
                    format_section(
                        "comments", submission.submission_comments, json_format=True
                    )
                )
            else:
                body = ""
                for comment in submission.submission_comments:
                    body += f"{comment['author_name']} ({comment['created_at']}):\n\n"
                    body += comment["comment"] + "\n\n"
                formatted_submission += format_section("Comments", body)
    except AttributeError:
        pass
    if history:
        try:
            submission_history = submission.submission_history
        except AttributeError:
            pass
        else:
            if submission_history:
                versions = {}
                for version, prev_submission in enumerate(
                    submission.submission_history
                ):
                    version = str(version)
                    version_dir = tmpdir / f"version-{version}"

                    prev_submission = canvasapi.submission.Submission(
                        submission._requester, prev_submission
                    )
                    prev_submission.assignment = submission.assignment

                    prev_metadata = format_submission(
                        prev_submission,
                        tmpdir=version_dir,
                        json_format=json_format,
                        md_title_level=md_title_level + "#",
                    )

                    versions[version] = prev_metadata
                    if json_format:
                        with open(version_dir / "metadata.json", "w") as f:
                            json.dump(prev_metadata, f, indent=2)
                    else:
                        with open(version_dir / "metadata.md", "w") as f:
                            f.write(prev_metadata)

                if json_format:
                    formatted_submission.update(
                        format_section(
                            "submission_history",
                            versions,
                            json_format=True,
                            md_title_level=md_title_level,
                        )
                    )
                else:
                    formatted_versions = ""
                    for version, prev_metadata in versions.items():
                        formatted_versions += format_section(
                            f"Version {version}",
                            prev_metadata,
                            md_title_level=md_title_level + "#",
                        )
                    formatted_submission += format_section(
                        "Submission history",
                        formatted_versions,
                        md_title_level=md_title_level,
                    )
    else:
        try:
            if submission.body:
                if json_format:
                    formatted_submission.update(
                        format_section(
                            "body",
                            submission.body,
                            json_format=True,
                            md_title_level=md_title_level,
                        )
                    )
                else:
                    formatted_submission += format_section(
                        "Body", submission.body, md_title_level=md_title_level
                    )
        except AttributeError:
            pass
        try:
            if submission.submission_data:
                if json_format:
                    formatted_submission.update(
                        format_section(
                            "quiz_answers",
                            submission.submission_data,
                            json_format=True,
                            md_title_level=md_title_level,
                        )
                    )
                else:
                    formatted_submission += format_section(
                        "Quiz answers",
                        json.dumps(submission.submission_data, indent=2),
                        md_title_level=md_title_level,
                    )
        except AttributeError:
            pass
        try:
            tmpdir = pathlib.Path(tmpdir or tempfile.mkdtemp())
            tmpdir.mkdir(parents=True, exist_ok=True)
            if json_format:
                attachments = {}
            for attachment in submission.attachments:
                contents = convert_to_md(attachment, tmpdir)
                formatted_attachment = format_section(
                    attachment.filename,
                    contents,
                    json_format=json_format,
                    md_title_level=md_title_level + "#",
                )

                if json_format:
                    attachments.update(formatted_attachment)
                else:
                    formatted_submission += formatted_attachment

            if json_format and attachments:
                formatted_submission.update(
                    format_section(
                        "attachments",
                        attachments,
                        json_format=True,
                        md_title_level=md_title_level,
                    )
                )
        except AttributeError:
            pass

    return formatted_submission


def format_section(title, body, json_format=False, md_title_level="#"):
    """
    In the case of Markdown (default), we format the title as a header and the body
    as a paragraph. If we don't do JSON, but receive a dictionary as the body, we
    format it as a list of key-value pairs.

    `md_title_level` is the level of the title in Markdown, by default `#`.
    We'll use this to create a hierarchy of sections in the output.

    In the case of JSON, we return a dictionary with the title as the key and the
    body as the value.
    """
    if json_format:
        return {title: body}

    if isinstance(body, dict):
        return "\n".join(
            [
                f" - {key.capitalize().replace('_', ' ')}: {value}"
                for key, value in body.items()
            ]
        )

    return f"\n{md_title_level} {title}\n\n{body}\n\n"


def resolve_grader(submission):
    """
    Returns a user object if the submission was graded by a human.
    Otherwise returns None if ungraded or a descriptive string.
    """
    try:
        if submission.grader_id is None:
            return None
    except AttributeError:
        return None

    if submission.grader_id < 0:
        return "autograded"
    return submission.assignment.course.get_user(submission.grader_id)


def convert_to_md(attachment: canvasapi.file.File, tmpdir: pathlib.Path) -> str:
    """
    Converts `attachment` to Markdown. Returns the Markdown string.

    Store a file version in `tmpdir`.
    """
    outfile = tmpdir / attachment.filename
    attachment.download(outfile)
    content_type = getattr(attachment, "content-type")
    try:
        md_type = text_to_md(content_type)
        with open(outfile, "r") as f:
            contents = f.read()
        return f"```{md_type}\n{contents}\n```"
    except ValueError:
        pass
    try:
        return pypandoc.convert_file(outfile, "markdown")
    except Exception as err:
        return (
            f"Pandoc cannot convert this file. "
            f"The file is located at\n\n  {outfile}\n\n"
        )


def text_to_md(content_type):
    """
    Takes a text-based content type, returns Markdown code block type.
    Raises ValueError if not possible.
    """
    if content_type.startswith("text/"):
        content_type = content_type[len("text/") :]
    else:
        raise ValueError(f"Not text-based content type: {content_type}")

    if content_type.startswith("x-"):
        content_type = content_type[2:]
    if content_type == "python-script":
        content_type = "python"

    return content_type


def format_rubric(submission, json_format=False):
    """
    Format the rubric assessment of the `submission` in readable form.

    If `json_format` is True, return a JSON string, otherwise Markdown.
    """

    if json_format:
        result = {}
    else:
        result = ""

    for crit_id, rating_data in submission.rubric_assessment.items():
        criterion = get_criterion(crit_id, submission.assignment.rubric)
        rating = get_rating(rating_data["rating_id"], criterion)
        try:
            comments = rating_data["comments"]
        except KeyError:
            comments = ""

        if json_format:
            result[criterion["description"]] = {
                "rating": rating["description"] if rating else None,
                "points": rating["points"] if rating else None,
                "comments": comments,
            }
        else:
            result += f"{criterion['description']}: "
            if rating:
                result += f"{rating['description']} ({rating['points']})"
            else:
                result += "-"
            result += "\n"

        if not json_format:
            result += "\n"

    return result.strip()


def get_criterion(criterion_id, rubric):
    """Returns criterion with ID `criterion_id` from rubric `rubric`"""
    for criterion in rubric:
        if criterion["id"] == criterion_id:
            return criterion

    return None


def get_rating(rating_id, criterion):
    """Returns rating with ID `rating_id` from rubric criterion `criterion`"""
    for rating in criterion["ratings"]:
        if rating["id"] == rating_id:
            return rating

    return None


def add_command(subp):
    """Adds the submissions and submission commands to argparse parser subp"""
    add_submissions_command(subp)
    add_submission_command(subp)


def add_submissions_command(subp):
    """Adds submissions command to argparse parser subp"""
    submissions_parser = subp.add_parser(
        "submissions",
        help="Lists submissions of an assignment",
        description="Lists submissions of assignment(s). Output format: "
        "<course code> <assignment name> <user> <grade> "
        "<submission date> <grade date>",
    )
    submissions_parser.set_defaults(func=submissions_command)
    assignments.add_assignment_option(submissions_parser)
    add_submission_options(submissions_parser)
    submissions_parser.add_argument(
        "-l",
        "--login-id",
        help="Print login ID instead of name.",
        default=False,
        action="store_true",
    )


def add_submission_command(subp):
    """Adds submission command to argparse parser subp"""
    submission_parser = subp.add_parser(
        "submission",
        help="Prints information about a submission",
        description="""
  Prints data about matching submissions, including submission and grading time, 
  and any text-based attachments.

  Uses MANPAGER or PAGER environment variables for the pager to page output. If 
  the `-r` or `-R` flag is passed to `less`, it uses colours in the output.

  """,
    )
    submission_parser.set_defaults(func=submission_command)
    add_submission_options(submission_parser)
    submission_parser.add_argument(
        "-o",
        "--output-dir",
        required=False,
        default=None,
        help="Write output to files in directory the given directory. "
        "If not specified, print to stdout. "
        "If specified, do not print to stdout.",
    )
    submission_parser.add_argument(
        "--json",
        required=False,
        action="store_true",
        default=False,
        help="Print output as JSON, otherwise Markdown.",
    )
    submission_parser.add_argument(
        "-H", "--history", action="store_true", help="Include submission history."
    )
    submission_parser.add_argument(
        "--open",
        required=False,
        nargs="?",
        default=None,
        const="open",
        choices=["open", "shell"],
        help="Open the directory containing the files using "
        "the default file manager (`open`) or "
        "spawn a shell in the directory (`shell`). "
        "With `open`, the pager will be used to display the output as usual. "
        "With `shell`, we just drop into the shell, the output can be "
        "found in the metadata.{json,md} file in the shell's working "
        "directory. "
        "Default: %(default)s",
    )
    submission_parser.add_argument(
        "--sort-order",
        required=False,
        choices=["student", "course"],
        default="student",
        help="Determines the order in which directories are created "
        "in `output_dir`. `student` results in `student/course/assignment` "
        "and `course` results in `course/assignment/student`. "
        "Default: %(default)s",
    )



import canvaslms.cli
from canvaslms.cli import assignments, courses, submissions, users
import canvaslms.hacks.canvasapi

import argparse
import csv
import canvasapi.submission
import datetime as dt
import importlib
import importlib.machinery
import importlib.util
import os
import pathlib
import re
import sys


def results_command(config, canvas, args):
    output = csv.writer(sys.stdout, delimiter=args.delimiter)

    if args.assignment_group != "":
        results = summarize_assignment_groups(canvas, args)
    else:
        results = summarize_assignments(canvas, args)

    for result in results:
        output.writerow(result)


def filter_grade(grade, regex):
    """
    Returns True if the grade matches the regex.
    """
    return re.search(regex, grade)


def summarize_assignments(canvas, args):
    """
    Turn submissions into results:
    - canvas is a Canvas object,
    - args is the command-line arguments, as parsed by argparse.
    """

    assignments_list = assignments.process_assignment_option(canvas, args)
    users_list = users.process_user_or_group_option(canvas, args)

    submissions_list = submissions.filter_submissions(
        submissions.list_submissions(assignments_list, include=["submission_history"]),
        users_list,
    )

    for submission in submissions_list:
        if submission.grade is not None:
            if filter_grade(submission.grade, args.filter_grades):
                yield [
                    submission.assignment.course.course_code,
                    submission.assignment.name,
                    submission.user.integration_id,
                    submission.grade,
                    round_to_day(submission.submitted_at or submission.graded_at),
                    *all_graders(submission),
                ]


def round_to_day(timestamp):
    """
    Takes a Canvas timestamp and returns the corresponding datetime.date object.
    """
    return dt.date.fromisoformat(timestamp.split("T")[0])


def all_graders(submission):
    """
    Returns a list of everyone who participated in the grading of the submission.
    I.e. also those who graded previous submissions, when submission history is
    available.
    """
    graders = []

    for prev_submission in submission.submission_history:
        prev_submission = canvasapi.submission.Submission(
            submission._requester, prev_submission
        )
        prev_submission.assignment = submission.assignment
        grader = submissions.resolve_grader(prev_submission)
        if grader:
            graders.append(grader)

    return graders


def summarize_assignment_groups(canvas, args):
    """
    Summarize assignment groups into a single grade:
    - canvas is a Canvas object,
    - args is the command-line arguments, as parsed by argparse.
    """

    courses_list = courses.process_course_option(canvas, args)
    all_assignments = list(assignments.process_assignment_option(canvas, args))
    users_list = set(users.process_user_or_group_option(canvas, args))

    for course in courses_list:
        ag_list = assignments.filter_assignment_groups(course, args.assignment_group)

        for assignment_group in ag_list:
            assignments_list = list(
                assignments.filter_assignments_by_group(
                    assignment_group, all_assignments
                )
            )
            if args.missing:
                if args.missing:
                    try:
                        missing = load_module(args.missing)
                    except Exception as err:
                        canvaslms.cli.err(
                            1,
                            f"Error loading missing module " f"'{args.missing}': {err}",
                        )
                if missing.missing_assignments == missing_assignments:
                    missing_results = missing.missing_assignments(
                        assignments_list,
                        users_list,
                        passing_regex=args.filter_grades,
                        optional_assignments=args.optional_assignments,
                    )
                else:
                    missing_results = missing.missing_assignments(
                        assignments_list, users_list
                    )
                for user, assignment, reason in missing_results:
                    yield [
                        course.course_code,
                        assignment_group.name,
                        user.login_id,
                        assignment.name,
                        reason,
                    ]
            else:
                try:
                    summary = load_module(args.summary_module)
                except Exception as err:
                    canvaslms.cli.err(
                        1,
                        f"Error loading summary module "
                        f"'{args.summary_module}': {err}",
                    )
                for user, grade, grade_date, *graders in summary.summarize_group(
                    assignments_list, users_list
                ):
                    if (
                        grade is None
                        or grade_date is None
                        or not filter_grade(grade, args.filter_grades)
                    ):
                        continue
                    yield [
                        course.course_code,
                        assignment_group.name,
                        user.integration_id,
                        grade,
                        grade_date,
                        *graders,
                    ]


def load_module(module_name):
    """
    Load a module from the file system or a built-in module.
    """
    try:
        return importlib.import_module(module_name)
    except ModuleNotFoundError:
        module_path = pathlib.Path.cwd() / module_name
        module = module_path.stem

        loader = importlib.machinery.SourceFileLoader(module, str(module_path))
        spec = importlib.util.spec_from_loader(module, loader)
        module_obj = importlib.util.module_from_spec(spec)
        loader.exec_module(module_obj)
        return module_obj


def missing_assignments(
    assignments_list,
    users_list,
    passing_regex=r"^([A-EP]|complete)$",
    optional_assignments=None,
):
    """
    Returns tuples of missing assignments.

    For each assignment that a student is not done with, we yield a tuple of the
    user, the assignment and the reason why the assignment is missing.

    The reason can be "not submitted" or "not graded" or "not a passing grade".

    The only reason to use a different module is if you have optional assignments.
    We only want to remind the students of the things they need to pass the course.
    We don't want to make it sound like an optional assignment is mandatory.
    """
    for user in users_list:
        for assignment in assignments_list:
            if optional_assignments:
                if any(
                    re.search(optional, assignment.name)
                    for optional in optional_assignments
                ):
                    continue
            try:
                submission = assignment.get_submission(user)
            except canvasapi.exceptions.ResourceDoesNotExist:
                continue

            if submission is None:
                yield user, assignment, "not submitted"
            elif submission.grade is None:
                if submission.submitted_at:
                    yield user, assignment, f"submitted on {submission.submitted_at}, but not graded"
                else:
                    yield user, assignment, "not done"
            elif not filter_grade(submission.grade, passing_regex):
                if (
                    submission.submitted_at
                    and submission.submitted_at > submission.graded_at
                ):
                    yield user, assignment, f"not a passing grade ({submission.grade}), resubmission not graded"
                else:
                    yield user, assignment, f"not a passing grade ({submission.grade})"


def add_command(subp):
    """Adds the results command to argparse parser subp"""
    results_parser = subp.add_parser(
        "results",
        help="Lists results of a course",
        description="""Lists results of a course for export, for instance to the `ladok report` 
                     command. Output format, CSV:

                       <course code> <component code> <student ID> <grade> <grade date> <graders ...>

                     Can also export a list of missing assignment results (--missing option) that 
                     prevent the student from getting a grade. Output format, CSV:

                       <course code> <component code> <student ID> <missing assignment> <reason>

                     The reason can be "not submitted" or "not graded".""",
        epilog="""If you specify an assignment group, the results of the assignments in that 
                group will be summarized. You can supply your own function for summarizing 
                grades through the -S option. See `pydoc3 canvaslms.grades` for different 
                options.""",
    )
    results_parser.set_defaults(func=results_command)
    assignments.add_assignment_option(results_parser, ungraded=False)
    users.add_user_or_group_option(results_parser)
    passing_regex = r"^([A-EP]|complete)$"
    all_grades_regex = r"^([A-F]x?|(in)?complete)$"
    results_parser.add_argument(
        "-F",
        "--filter-grades",
        required=False,
        action="store",
        nargs="?",
        const=all_grades_regex,
        default=passing_regex,
        help=f"Filter grades. By default we only output "
        f"A--Es and Ps ({passing_regex}. "
        f"If you want to include Fs ({all_grades_regex}), use this option. "
        f"You can also supply an optional regex to this option "
        f"to filter grades based on that.",
    )
    default_summary_module = "canvaslms.grades.conjunctavg"
    results_parser.add_argument(
        "-S",
        "--summary-module",
        required=False,
        default=default_summary_module,
        help=f"""Name of Python module or file containing module to load with a custom 
             summarization function to summarize assignment groups. The default module is 
             part of the `canvaslms` package: `{default_summary_module}`. But it could be 
             any Python file in the file system or other built-in modules. See `pydoc3 
             canvaslms.grades` for alternative modules or how to build your own.""",
    )
    default_missing_module = "canvaslms.cli.results"
    results_parser.add_argument(
        "-M",
        "--missing",
        required=False,
        nargs="?",
        const=default_missing_module,
        default=None,
        help="Produce a list of missing assignments instead of grades. "
        "You can supply a custom module to this option, the module must "
        "contain a "
        "function `missing_assignments(assignments_list, users_list). "
        "The default module checks if all things are graded or submitted. "
        "This option only has effect when working with assignment groups.",
    )
    results_parser.add_argument(
        "-O",
        "--optional-assignments",
        required=False,
        nargs="+",
        default=None,
        help="List of regexes matching optional assignments. The default missing "
        "assignments will treat matching assignments as optional.",
    )
